//
// Vertex shader for procedural bricks
//
// Authors: Dave Baldwin, Steve Koren, Randi Rost
//          based on a shader by Darwyn Peachey
//
// Copyright (c) 2002-2004 3Dlabs Inc. Ltd. 
//
// See 3Dlabs-License.txt for license information
//

const float ambientContribution = 0.2;  // added for ambient light
const float SpecularContribution = 0.3;
const float DiffuseContribution  = 1.0 - SpecularContribution - ambientContribution;

varying float LightIntensity;
varying vec2  MCposition;
varying int   NormDir;

void main(void)
{
    vec3 ecPosition = vec3 (gl_ModelViewMatrix * gl_Vertex);
    vec3 tnorm      = normalize(gl_NormalMatrix * gl_Normal);
    vec3 lightVec   = normalize(vec3(gl_LightSource[0].position) - ecPosition);
    vec3 reflectVec = reflect(-lightVec, tnorm);
    vec3 viewVec    = normalize(-ecPosition);
    float diffuse   = max(dot(lightVec, tnorm), 0.0);
    float spec      = 0.0;
	 // calculate tangential vector of vertex normal
	 vec2 tanVec     = normalize(vec2(gl_Normal.z, -gl_Normal.x));
	 // need absolute value of x and z component to classify surfaces
	 vec2 absNormal  = abs(gl_Normal.xz);
	 
    if (diffuse > 0.0)
    {
        spec = max(dot(reflectVec, viewVec), 0.0);
        spec = pow(spec, 16.0);
    }

    LightIntensity  = DiffuseContribution * diffuse +
                      SpecularContribution * spec + ambientContribution;
							 
	 // horizontal texture coordinate is the projection of (x, z) to the tangential surface
    MCposition      = vec2(dot(gl_Vertex.xz, tanVec), gl_Vertex.y);

	 // classifying surfaces
	 if (absNormal[0] < absNormal[1])
	     NormDir = 1;
	 else
        NormDir = 0;
		  	  
	 gl_Position     = ftransform();
}
