// Fragment shader for environment mapping with an
// equirectangular 2D texture
// Authors: John Kessenich, Randi Rost
// Copyright (c) 2002-2004 3Dlabs Inc. Ltd.
// See 3Dlabs-License.txt for license information

uniform vec3 BaseColor;
uniform float MixRatio;
uniform sampler2D envMap;

varying vec3 Normal;
varying vec3 EyeDir;
varying float LightIntensity;

void main (void)
{
    // compute reflection vector
    vec3 reflectDir = normalize(reflect(EyeDir, Normal));
	
    // compute altitude and azimuth angles
    vec2 index;

	 // because the row order of Jpeg image and OpenGL texture are opposite
    index.y = 1 - (reflectDir.y + 1) * 0.5; 
    index.x = (reflectDir.x + 1) * 0.5;

    // Do a lookup into the environment map
	 vec3 envColor = vec3 (texture2D(envMap, index));

    // Add lighting to base colour and mix
    vec3 base = LightIntensity * BaseColor;
    envColor = mix(base, envColor, MixRatio);

    gl_FragColor = vec4(envColor, 1.0);
}
