/* ===========================================================================
Copyright (C) 2002-2005  3Dlabs Inc. Ltd.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

    Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

    Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

    Neither the name of 3Dlabs Inc. Ltd. nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
=========================================================================== */

//
// The source of the shaders itself are in plain text files. Files ending in
// .vert are vertex shaders, and files ending in .frag are fragment shaders.
//

#include "ogl2brick.h"
#include <wand/magick_wand.h>

//
// Handle of the window we're rendering to
//
static GLint window;
GLuint brickProg, toonProg, envProg;

//
// Models to render
//
EModelType gleModel = EModelCube;

//
// Movement variables
//
float fXDiff = 0;
float fYDiff = 0;
float fZDiff = 0;
int xLastIncr = 0;
int yLastIncr = 0;
float fXInertia = -0.5;
float fYInertia = 0;
float fXInertiaOld;
float fYInertiaOld;
float fScale = 1.0;
float ftime = 0;
int xLast = -1;
int yLast = -1;
char bmModifiers;
int Rotate = 1;
float lpos[3] = {-5.0, 1.0, 5.0};
float aspect = 1.0;
float mixRatio = 0.5;
int textureID = 0;

//
// Rotation defines
//
#define INERTIA_THRESHOLD       1.0f
#define INERTIA_FACTOR          0.5f
#define SCALE_FACTOR            0.01f
#define SCALE_INCREMENT         0.1f
#define TIMER_FREQUENCY_MILLIS  20

GLfloat RotL = 1 * 3.14f / 180;
int LastTime = 0;

/*public*/
int printOglError(char *file, int line)
{
    //
    // Returns 1 if an OpenGL error occurred, 0 otherwise.
    //
    GLenum glErr;
    int    retCode = 0;

    glErr = glGetError();
    while (glErr != GL_NO_ERROR)
    {
        printf("glError in file %s @ line %d: %s\n", file, line, gluErrorString(glErr));
        retCode = 1;
        glErr = glGetError();
    }
    return retCode;
}


/***************************************************************************/
/* Parse GL_VERSION and return the major and minor numbers in the supplied
 * integers.
 * If it fails for any reason, major and minor will be set to 0.
 * Assumes a valid OpenGL context.
*/

void getGlVersion( int *major, int *minor )
{
    const char* verstr = (const char*)glGetString( GL_VERSION );
    if( (verstr == NULL) || (sscanf( verstr, "%d.%d", major, minor ) != 2) )
    {
        *major = *minor = 0;
        fprintf( stderr, "Invalid GL_VERSION format!!!\n" );
    }
}


static void NextClearColor(void)
{
    static int color = 0;

    switch( color++ )
    {
        case 0:  glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
             break;
        case 1:  glClearColor(0.2f, 0.2f, 0.3f, 1.0f);
             break;
        default: glClearColor(0.7f, 0.7f, 0.7f, 1.0f);
             color = 0;
             break;
    }
}



/******************************************************************************

 GLUT glue

******************************************************************************/

static
void display(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	// set perspective projection
	glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
	gluPerspective(60.0, aspect, 1, 100);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

	// using OpenGL - disable shaders
	glUseProgram(0);
	glBindTexture(GL_TEXTURE_2D, textureID);

	glPushMatrix();
	// set camera for background
	gluLookAt(0.0,0.0,1.0, 0.0,0.0,-1.0, 0.0f,1.0f,0.0f);
	glTranslatef(0, 0, -10);
	glScalef(10.0, 10.0, 0.0);

	// draw background
	glBegin( GL_POLYGON );
		glTexCoord2f( 0.0f, 1.0f );
		glVertex3f( -1.0f, -1.0f, 0.0f);

		glTexCoord2f( 1.0f, 1.0f );
		glVertex3f(  1.0f, -1.0f, 0.0f);

		glTexCoord2f( 1.0f, 0.0f );        
		glVertex3f(  1.0f,  1.0f, 0.0f);

		glTexCoord2f( 0.0f, 0.0f );     
		glVertex3f( -1.0f,  1.0f, 0.0f);
	glEnd();    
	glPopMatrix();

	//set camera for objects
   gluLookAt(0.0,2.5,4.0, 0.0,0.0,-1.0, 0.0f,1.0f,0.0f);
				  
	//set light position			  
	glLightfv(GL_LIGHT0, GL_POSITION, lpos);
	 
	 glRotatef(fYDiff, 1,0,0);
    glRotatef(fXDiff, 0,1,0);
    glRotatef(fZDiff, 0,0,1);

    glScalef(fScale, fScale, fScale);
	
	// draw bricks
	 glUseProgram(brickProg);
	 glutSolidCube(2.0f);

	// draw teapot
	 glUseProgram(toonProg);
 	 glUniform1f(getUniLoc(toonProg,"MixRatio"), mixRatio);
	 glPushMatrix();
 	 glTranslatef(0.0f, 1.6f, 0.0f);
	 glutSolidTeapot(0.8f);
    glPopMatrix();
	 
	// draw bottom surface 
	 glUseProgram(envProg);
 	 glUniform1f(getUniLoc(envProg,"MixRatio"), mixRatio);
 	 glTranslatef(0.0f, -1.1f, 0.0f);
	 glScalef(8.0f, 0.1f, 8.0f);
	 glutSolidCube(1.0f);

	 glFlush();
    glutSwapBuffers();
}


static
void play(void)
{
    int thisTime = glutGet(GLUT_ELAPSED_TIME);

    RotL += (thisTime - LastTime) * -0.001f;
    LastTime = thisTime;

    glutPostRedisplay();
}


static
void key(unsigned char keyPressed, int x, int y)
{
    switch(keyPressed)
    {
        case 'b':
            NextClearColor();
            break;
        case 'q':
        case 27:
            exit(0);
            break;
        case 't':
            textureID = (textureID + 1) % 3;
            break;
        case ' ':
            Rotate = !Rotate;

            if (!Rotate)
            {
                fXInertiaOld = fXInertia;
                fYInertiaOld = fYInertia;
            }
            else
            {
                fXInertia = fXInertiaOld;
                fYInertia = fYInertiaOld;

                // To prevent confusion, force some rotation
                if ((fXInertia == 0.0) && (fYInertia == 0.0))
                    fXInertia = -0.5;
            }
            break;
        case '+':
            fScale += SCALE_INCREMENT;
            break;
        case '-':
            fScale -= SCALE_INCREMENT;
            break;
		  case 'o':
            mixRatio -= 0.05;
            break;				
		  case 'p':
            mixRatio += 0.05;
            break;				
        default:
            fprintf(stderr, "\nKeyboard commands:\n\n"
            "b - Toggle among background clear colors\n"
            "q, <esc> - Quit\n"
            "t - Toggle among models to render\n"
            "? - Help\n"
            "<home>     - reset zoom and rotation\n"
            "<space> or <click>        - stop rotation\n"
            "<+>, <-> or <ctrl + drag> - zoom model\n"
            "<arrow keys> or <drag>    - rotate model\n\n");
            break;
    }
}


static
void timer(int value)
{
    /* Increment timer */
    ftime += 0.01f;

    /* Increment wrt inertia */
    if (Rotate)
    {
        fXDiff = (fXDiff + fXInertia);
        fYDiff = (fYDiff + fYInertia);
    }

    /* Callback */
    glutTimerFunc(TIMER_FREQUENCY_MILLIS , timer, 0);
}


static
void mouse(int button, int state, int x, int y)
{
   bmModifiers = glutGetModifiers();

   if (button == GLUT_LEFT_BUTTON)
   {
      if (state == GLUT_UP)
      {
         xLast = -1;
         yLast = -1;

         if (xLastIncr > INERTIA_THRESHOLD)
            fXInertia = (xLastIncr - INERTIA_THRESHOLD)*INERTIA_FACTOR;

         if (-xLastIncr > INERTIA_THRESHOLD) 
            fXInertia = (xLastIncr + INERTIA_THRESHOLD)*INERTIA_FACTOR;

         if (yLastIncr > INERTIA_THRESHOLD) 
            fYInertia = (yLastIncr - INERTIA_THRESHOLD)*INERTIA_FACTOR;

         if (-yLastIncr > INERTIA_THRESHOLD) 
            fYInertia = (yLastIncr + INERTIA_THRESHOLD)*INERTIA_FACTOR;
      }
      else
      {
         fXInertia = 0;
         fYInertia = 0;
      }
      xLastIncr = 0;
      yLastIncr = 0;
   }
}


static
void motion(int x, int y)
{
   if ((xLast != -1) || (yLast != -1))
   {
      xLastIncr = x - xLast;
      yLastIncr = y - yLast;
      if (bmModifiers & GLUT_ACTIVE_CTRL)
      {
         if (xLast != -1)
         {
            fZDiff += xLastIncr;
            fScale += (yLastIncr)*SCALE_FACTOR;
         }
      }
      else
      {
         if (xLast != -1)
         {
            fXDiff += xLastIncr;
            fYDiff += yLastIncr;
         }
      }
   }

   xLast = x;
   yLast = y;
}


static
void reshape(int wid, int ht)
{
    aspect = (float) wid / (float) ht;

    glViewport(0, 0, wid, ht);
}


static
void special(int key, int x, int y)
{
    switch (key)
    {
        case GLUT_KEY_HOME:
            fXDiff    = 0;
            fYDiff    = 0;
            fZDiff    = 0;
            xLastIncr = 0;
            yLastIncr = 0;
            fXInertia = -0.5;
            fYInertia = 0;
            fScale    = 1.0;
				lpos[0]   = -5.0;
				lpos[1]   = 1.0;
				lpos[2]   = 5.0;
        break;
		  
		  // change light position
        case GLUT_KEY_LEFT:
           lpos[0] = lpos[0] - 0.1;
        break;
        case GLUT_KEY_RIGHT:
           lpos[0] = lpos[0] + 0.1;
        break;
        case GLUT_KEY_UP:
           lpos[1] = lpos[1] + 0.1;
        break;
        case GLUT_KEY_DOWN:
           lpos[1] = lpos[1] - 0.1;
        break;
        case GLUT_KEY_PAGE_UP:
           lpos[2] = lpos[2] - 0.1;
        break;
        case GLUT_KEY_PAGE_DOWN:
           lpos[2] = lpos[2] + 0.1;
        break;

    }
}

// read an image using ImageMagick and load it to texture
int loadTexture(char *fileName, int texID)
{
	MagickBooleanType status;  
	MagickWand *magick_wand;
	
	magick_wand=NewMagickWand();  
	status = MagickReadImage(magick_wand, fileName);

	if (status == MagickFalse) {
		printf("Fail to read image %s\n", fileName);
		return 1;
	}
		
	int width = MagickGetImageWidth(magick_wand);
	int height = MagickGetImageHeight(magick_wand);

	unsigned char *pixels = (unsigned char*)malloc(width*height*3);
	//memset(pixels, 0, width*height*3);

	MagickGetImagePixels(magick_wand, 0, 0, width, height, "RGB", CharPixel, pixels);
	DestroyMagickWand(magick_wand);
	
	glBindTexture(GL_TEXTURE_2D, texID);
	
   glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGB,
        GL_UNSIGNED_BYTE, (const GLvoid *) pixels);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	
	return 0;
}

/*public*/
int installBrickShaders(const GLchar *brickVertex,
                        const GLchar *brickFragment)
{
	 brickProg = installShaders(brickVertex, brickFragment); // Program handle

    if (!brickProg)
        return 0;

    // Install program object as part of current state

    glUseProgram(brickProg);

    // Set up initial uniform values

    glUniform3f(getUniLoc(brickProg, "BrickColor"), 1.0, 0.3, 0.2);
    glUniform3f(getUniLoc(brickProg, "MortarColor"), 0.85, 0.86, 0.84);
    glUniform2f(getUniLoc(brickProg, "BrickSize"), 0.30, 0.15);
    glUniform2f(getUniLoc(brickProg, "BrickPct"), 0.90, 0.85);
	 glUniform1i(getUniLoc(brickProg, "rowNum"), (int)floor(1.0/0.15));

    return 1;
}

int installToonShaders(const GLchar *toonVertex,
                        const GLchar *toonFragment)
{
	toonProg = installShaders(toonVertex, toonFragment); // Program handle

   if (!toonProg)
      return 0;

   // Install program object as part of current state

   glUseProgram(toonProg);

    // Set up initial uniform values	 
	GLint loc1,loc2,loc3,loc4,loc5,loc6;
	float specIntensity = 0.98;
	float sc[4] = {0.8,0.8,0.8,1.0};
	float threshold[2] = {0.5,0.25};
	float colors[12] = {0.6,0.6,0.6,1.0,
						0.3,0.3,0.3,1.0,
						0.1,0.1,0.1,1.0};
	
	loc1 = getUniLoc(toonProg,"specIntensity");
	glUniform1f(loc1,specIntensity);

	loc2 = getUniLoc(toonProg,"specColor");
   glUniform4fv(loc2,1,sc);
    
	loc3 = getUniLoc(toonProg,"t");
	glUniform1fv(loc3,2,threshold);

	loc4 = getUniLoc(toonProg,"colors");
	glUniform4fv(loc4,3,colors);
	
	loc5 = getUniLoc(toonProg,"MixRatio");
   glUniform1f(loc5, mixRatio);
    
	loc6 = getUniLoc(toonProg,"envMap");
	glUniform1i(loc6, 0);  // texture ID = 0
	
   return 1;
}

int installEnvShaders(const GLchar *envVertex,
                        const GLchar *envFragment)
{
	envProg = installShaders(envVertex, envFragment); // Program handle

   if (!envProg)
      return 0;

   // Install program object as part of current state

   glUseProgram(envProg);

    // Set up initial uniform values	 
	GLint loc1,loc2,loc3;
	float baseColor[3] = {0.6,0.6,0.6};
	
	loc1 = getUniLoc(envProg,"BaseColor");
	glUniform3fv(loc1, 1, baseColor);

	loc2 = getUniLoc(envProg,"MixRatio");
   glUniform1f(loc2, mixRatio);
    
	loc3 = getUniLoc(envProg,"envMap");
	glUniform1i(loc3, 0);  // texture ID = 0

   return 1;
}


/******************************************************************************

 Main

******************************************************************************/

int main( int argc, char **argv )
{
    int gl_major, gl_minor;
    GLchar *VertexShaderSource, *FragmentShaderSource;

    glutInit( &argc, argv );
    glutInitDisplayMode( GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
    glutInitWindowSize(500, 500);
    window = glutCreateWindow( "CS594 - Proj 1");

    glutIdleFunc(play);
    glutDisplayFunc(display);
    glutKeyboardFunc(key);
    glutReshapeFunc(reshape);
    glutMotionFunc(motion);
    glutMouseFunc(mouse);
    glutSpecialFunc(special);
    glutTimerFunc(TIMER_FREQUENCY_MILLIS , timer, 0);

    // Initialize the "OpenGL Extension Wrangler" library
    glewInit();

    // Make sure that OpenGL 2.0 is supported by the driver
    getGlVersion(&gl_major, &gl_minor);
    printf("GL_VERSION major=%d minor=%d\n", gl_major, gl_minor);

    if (gl_major < 2)
    {
        printf("GL_VERSION major=%d minor=%d\n", gl_major, gl_minor);
        printf("Support for OpenGL 2.0 is required for this demo...exiting\n");
        exit(1);
    }

    glDepthFunc(GL_LESS);
    glEnable(GL_DEPTH_TEST);
	 glEnable(GL_TEXTURE_2D); // enable 2D texture
    
	 NextClearColor();

    key('?', 0, 0);     // display help

	// load shaders
    readShaderSource("brick", &VertexShaderSource, &FragmentShaderSource);
    if (!installBrickShaders(VertexShaderSource, FragmentShaderSource))
		 return 0;

    readShaderSource("toon", &VertexShaderSource, &FragmentShaderSource);
    if (!installToonShaders(VertexShaderSource, FragmentShaderSource))
		 return 0;
	 
    readShaderSource("env", &VertexShaderSource, &FragmentShaderSource);
    if (!installEnvShaders(VertexShaderSource, FragmentShaderSource))
		 return 0;
		 
	// load environment texture	
	 loadTexture("landscape.jpg", 0);
	 loadTexture("env.jpg", 1);
	 loadTexture("ss204.jpg", 2);
    
    glutMainLoop();
    
    return 0;
}
