uniform float specIntensity;
uniform vec4  specColor;
uniform float t[2];
uniform vec4  colors[3];
uniform float MixRatio;
uniform sampler2D envMap;

// all of the uniform variables could have been defined as liters within the shader
// but this way we get to pass those values in from the OpenGL program

varying vec3  normal, lightDir, EyeDir;

void main()
{
    float intensity;
    vec3 n;
    vec4 color;

    n = normalize(normal);
    intensity = max(dot(lightDir,n),0.0);

    if (intensity > specIntensity)
        color = specColor;
    else if (intensity > t[0])
        color = colors[0];   
    else if (intensity > t[1])
        color = colors[1];
    else
        color = colors[2];       

	// compute reflection vector
    vec3 reflectDir = normalize(reflect(EyeDir, normal));
	
    // compute altitude and azimuth angles
    vec2 index;

    index.y = 1 - (reflectDir.y + 1) * 0.5;
    index.x = (reflectDir.x + 1) * 0.5;

    // Do a lookup into the environment map
    vec4 envColor = vec4(texture2D(envMap, index));
	 color = mix(color, envColor, MixRatio);

    gl_FragColor = color;
}
